import base64
import functools
import logging
from http.client import HTTPResponse

from .data import GallonModel, dumps
from .objects import Response


@functools.singledispatch
def make_response(response: Response):
    """Return the response"""
    return response


@make_response.register(HTTPResponse)
def _(response: HTTPResponse):
    """Return the response"""
    try:
        data = response.read().decode("utf-8")
    except:
        data = base64.b64encode(response.read()).decode(
            "utf-8"
        )  # pylint: disable=bare-except
    return Response(
        status=response.status,
        headers=dict(response.getheaders()),
        body=data,
    )


@make_response.register(Response)
def _(response: Response):
    """Return the response"""
    return response


@make_response.register(str)
def _(response: str):
    if response.startswith("<"):
        return make_response(
            Response(
                status=200,
                headers={"Content-Type": "text/html"},
                body=response,
            )
        )
    return make_response(
        Response(
            status=200,
            headers={"Content-Type": "text/plain"},
            body=response,
        )
    )


@make_response.register(dict)
def _(response: dict):
    return make_response(
        Response(
            status=200,
            headers={"Content-Type": "application/json"},
            body=dumps(response),
        )
    )


@make_response.register(GallonModel)
def _(response: GallonModel):
    return make_response(
        Response(
            status=200,
            headers={"Content-Type": "application/json"},
            body=dumps(response.dict()),
        )
    )


@make_response.register(list)
def _(response: list):
    _response = [dumps(r) for r in response]
    return make_response(
        Response(
            status=200,
            headers={"Content-Type": "application/json"},
            body=dumps(_response),
        )
    )

def setup_logger():
    """Setup the logger"""
    logger = logging.getLogger("Gallon Server")
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    logger.addHandler(handler)
    return logger


"""HTTP client"""
from __future__ import annotations

import http.client
import http.server
import json
import urllib.parse
from typing import Any, Dict, Optional

from .data import dumps
from .decorators import asyncify
from .objects import Request, Response


class GallonClient:
    """A simple HTTP client"""

    def fetch(self, request: Request):
        """Send a request"""
        parsed_url = urllib.parse.urlparse(request.url)
        conn = self._get_connection(parsed_url)
        headers = self._prepare_headers(request.headers)
        body = self._prepare_body(request)
        if body is None:
            conn.request(request.method, parsed_url.path, headers=headers)
        else:
            if request.method in ("GET", "HEAD"):
                raise ValueError(f"Request method cannot have a body: {request.method}")   
            conn.request(request.method, parsed_url.path, body.encode(), headers)

        with conn.getresponse() as res:
            response_headers = self._parse_response_headers(res)
            response_body = self._parse_response_body(res, response_headers)
            response = Response(
                status=res.status, headers=response_headers, body=response_body
            )
            return response.dict()

    def _prepare_body(self, request: Request):
        """Prepare the body based on the content type"""
        _json = request.json()
        if _json is None:
            return None
        if request.content_type == "application/x-www-form-urlencoded":
            assert isinstance(_json, dict)
            return urllib.parse.urlencode(_json)
        if request.content_type == "application/json":
            return dumps(_json)
        return request.body
        
    def _url_encode_body(self, body: Dict[str, str]) -> str:
        """Encode the body as a URL-encoded string"""
        encoded_params = urllib.parse.urlencode(body)
        return encoded_params

    def get(self, url: str, headers: Optional[Dict[str, str]] = None):
        """Send a GET request to a URL"""
        return self.fetch(Request(method="GET", url=url, body=None, headers=headers))
    

    def post(
        self, url: str, body: Dict[str, Any], headers: Optional[Dict[str, str]] = None
    ):
        """Send a POST request to a URL"""
        return self.fetch(Request(method="POST", url=url, body=body, headers=headers))

    def put(self, url: str, body: Any, headers: Optional[Dict[str, str]] = None):
        """Send a PUT request to a URL"""
        return self.fetch(Request(method="PUT", url=url, body=body, headers=headers))

    def delete(self, url: str, headers: Optional[Dict[str, str]] = None):
        """Send a DELETE request to a URL"""
        return self.fetch(
            Request(method="DELETE", url=url, body=None, headers=headers)
        )

    def _get_connection(self, parsed_url: urllib.parse.ParseResult):
        """Create an HTTPConnection or HTTPSConnection based on the URL scheme"""
        if parsed_url.scheme == "http":
            return http.client.HTTPConnection(parsed_url.netloc)
        elif parsed_url.scheme == "https":
            return http.client.HTTPSConnection(parsed_url.netloc)
        else:
            raise ValueError(f"Unsupported scheme: {parsed_url.scheme}")

    def _prepare_headers(self, headers: Dict[str, str]):
        """Prepare headers with defaults"""
        prepared_headers = {"User-Agent": "Gallon/0.1"}
        if headers:
            prepared_headers.update(headers)
        return prepared_headers

    def _parse_response_headers(self, response):
        """Parse response headers into a dictionary"""
        return {k.lower(): v for k, v in response.getheaders()}

    def _parse_response_body(self, response, headers):
        """Parse response body based on Content-Type header"""
        content_type = headers.get("content-type", "")
        if content_type.endswith("json"):
            return json.loads(response.read().decode())
        if content_type.startswith("text/"):
            return response.read().decode()
        if content_type.startswith("x-www-form-urlencoded"):
            return urllib.parse.parse_qs(response.read().decode())
        return response.read()



"""Gallon server"""
import functools
import logging
import os
import sys
import threading
import time
from http.server import HTTPServer, SimpleHTTPRequestHandler
from typing import Any, Callable, NamedTuple, Optional, cast

from .helpers import make_response, setup_logger
from .objects import Request, Response


class GallonRoute(NamedTuple):
    """Route data model"""

    method: str
    path: str


class GallonRouter:
    """A simple router"""

    def __init__(self):
        self.routes = {}
        self.middleware = []

    def route(self, method: str, path: str):
        """Register a route"""

        def decorator(handler: Callable):
            self.routes[(method, path)] = handler
            return handler

        return decorator

    def get_handler(self, method: str, path: str):
        """Return a handler for a route"""
        return self.routes.get((method, path))


class GallonHandler(SimpleHTTPRequestHandler):
    """HTTP request handler"""

    router = GallonRouter()
    
    def handle_request(self, method: str):
        """Handle a request"""
        handler = self.router.get_handler(method, self.path)
        if handler is not None:
            # Get middleware from route if it exists
            request = self.get_request(method)
            raw_response = handler(request)  # handler's return value
            response = make_response(
                raw_response
            )  # use make_response to format it properly
            self.send_response(response.status)
            for key, value in response.headers.items():
                self.send_header(key, value)
            self.send_header("Content-Length", str(len(response.body))) # type: ignore
            self.send_header("Access-Control-Allow-Origin", "*")
            self.send_header("Access-Control-Allow-Headers", "*")
            self.send_header("Access-Control-Allow-Methods", "*")
            self.send_header("Access-Control-Allow-Credentials", "true")
            self.send_header("Access-Control-Max-Age", "86400")
            self.end_headers()
            assert response.body is not None
            self.wfile.write(response.body.encode("utf-8"))
        else:
            self.send_response(404)
            self.end_headers()
            self.wfile.write(b"Not found")

    def get_request(self, method: str):
        """Return the request"""
        headers = {key: value for key, value in self.headers.items()}
        body = self.get_body()
        return Request(
            url=self.path,
            method=method,
            headers=headers,
            body=cast(Optional[str], body),
        )

    def get_body(self):
        """Return the body"""
        content_length = self.headers.get("Content-Length")
        if content_length is None:
            return None
        length = int(content_length)
        return self.rfile.read(length).decode("utf-8")

    def process_request(self, request: Request, handler: Callable):
        """Handle a request"""
        response = handler(request)
        return make_response(response)

    def do_GET(self):
        """Handle a GET request"""
        self.handle_request("GET")

    def do_POST(self):
        """Handle a POST request"""
        self.handle_request("POST")

    def do_PUT(self):
        """Handle a PUT request"""
        self.handle_request("PUT")

    def do_DELETE(self):
        """Handle a DELETE request"""
        self.handle_request("DELETE")

class Gallon(HTTPServer):
    """
    A Zero-Dependency Python Micro Framework
    Features:
    - Routing
    - Request and Response objects
    - JSON support
    - Data validation
    - Decorators
    - Static files
    """

    def __init__(self, host: str = "localhost", port: int = 8000):
        super().__init__((host, port), GallonHandler)
        self.handler = GallonHandler
        self.logger = logging.getLogger(__name__)
    
    def route(self, method: str, path: str):
        """Register a route"""
        def decorator(handler: Callable):
            self.handler.router.route(method, path)(handler)
            return handler
        return decorator

    
    def get(self, path: str):
        """Register a GET route"""
        return self.route("GET", path)
    
    def post(self, path: str):
        """Register a POST route"""
        return self.route("POST", path)
    def put(self, path: str):
        """Register a PUT route"""
        return self.route("PUT", path)

    def delete(self, path: str):
        """Register a DELETE route"""
        return self.route("DELETE", path)

    def run(self):
        """Run the server"""
        self.logger = setup_logger()
        self.logger.info("Starting server at http://%s:%s", *self.server_address)
        self.logger.info("Press Ctrl+C to stop the server")

        # Monitor for file changes in a separate thread
        self.file_change_monitor_thread = threading.Thread(target=self.monitor_files)
        self.file_change_monitor_thread.start()

        try:
            self.serve_forever()
        except KeyboardInterrupt:
            self.logger.info("Stopping server")
            self.shutdown()
            self.server_close()
            self.logger.info("Server stopped")
        finally:
            self.logger.handlers.clear()

    def monitor_files(self):
        """Monitor for changes in Python files and restart the server if a change is detected"""
        path_to_watch = "."  # Directory to monitor
        file_modification_times = self.track_files(path_to_watch)

        while True:
            time.sleep(1)
            new_file_modification_times = self.track_files(path_to_watch)
            if new_file_modification_times != file_modification_times:
                self.logger.info("Change detected, restarting server.")
                self.restart()
                break  # Exit the monitoring thread

    @staticmethod
    def track_files(path):
        """Get the modification times for all Python files in the specified path"""
        file_modification_times = {}
        for dirpath, dirnames, filenames in os.walk(path):
            for filename in filenames:
                if filename.endswith(".py"):
                    file_path = os.path.join(dirpath, filename)
                    file_modification_times[file_path] = os.path.getmtime(file_path)
        return file_modification_times

    def restart(self):
        """Restart the server"""
        self.logger.info("Restarting server")
        self.shutdown()
        self.server_close()
        os.execv(sys.executable, ['python'] + sys.argv)  # Start a new instance of the current script

        
        


"""
Flaskaesque helper functions for aiohttp.
"""
import asyncio
import functools
import types
import typing
from concurrent.futures import ThreadPoolExecutor


def asyncify(
    func: typing.Callable[..., typing.Any],
) -> typing.Callable[..., typing.Awaitable[typing.Any]]:
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if hasattr(self, "executor") is False:
            self.executor = ThreadPoolExecutor(max_workers=5)
        bound = functools.partial(func, self, *args, **kwargs)
        if asyncio.get_event_loop().is_running():
            loop = asyncio.get_event_loop()
            return loop.run_in_executor(self.executor, bound)
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop.run_in_executor(self.executor, bound)

    return typing.cast(typing.Callable[..., typing.Awaitable[typing.Any]], wrapper)


def aio(max_workers: int = 5) -> typing.Callable[[typing.Type], typing.Type]:
    """Decorator that converts all the methods of a class into async methods."""
    
    def decorator(cls: typing.Type) -> typing.Type:
        attrs: typing.Dict[str, typing.Any] = {}
        attrs["executor"] = ThreadPoolExecutor(max_workers=max_workers)
        for attr_name, attr_value in cls.__dict__.items():
            if (
                isinstance(attr_value, types.FunctionType)
                and attr_name.startswith("__") is False
            ):
                attrs[attr_name] = asyncify(attr_value)
            else:
                attrs[attr_name] = attr_value
        return type(cls.__name__, cls.__bases__, attrs)

    return decorator



"""Base Request and Response Objects"""
from typing import Any, Optional, cast
from urllib.parse import parse_qs, urlparse

from typing_extensions import override

from .data import GallonModel, Json, field, loads

JSON_CORS_HEADERS = {
    "Access-Control-Allow-Origin": "*",
    "Access-Control-Allow-Methods": "GET, POST, PUT, DELETE, OPTIONS",
    "Access-Control-Allow-Headers": "Content-Type",
    "Access-Control-Max-Age": "86400",
    "Content-Type": "application/json",
}

class Request(GallonModel):
    """
    Request Object
    Interface that defines the shape of incoming http requests
    """

    url: str = field(...)
    method: str = field(...)
    headers: dict = field(JSON_CORS_HEADERS)
    body: str = field(None)

    @override
    def json(self, exclude_none: bool = True):
        """
        Convert and return the body as a JSON object if Content-Type is application/json.

        Returns:
            None: if body is empty or Content-Type is not application/json.
            Json: the body as a JSON object.
        """
        if self.body is None:
            return None
        if self.headers.get("Content-Type") != "application/json":
            return None
        return cast(Json, loads(self.body))

    @property
    def content_type(self):
        """
        Get the value of the Content-Type header.

        Returns:
            str: the value of the Content-Type header.
        """
        return self.headers.get("Content-Type")

    @property
    def content_length(self):
        """
        Get the value of the Content-Length header.
        If Content-Length is not set, it calculates the length of the body and sets the header accordingly.

        Returns:
            str: the value of the Content-Length header.
        """
        if self.headers.get("Content-Length") is None:
            self.headers["Content-Length"] = str(len(self.body or ""))
        assert self.headers.get("Content-Length") == str(len(self.body or ""))
        return self.headers.get("Content-Length")

    @property
    def query(self):
        """
        Parse and return the query parameters from the URL.

        Returns:
            dict: the query parameters.
        """
        return {k: v[0] for k, v in parse_qs(urlparse(self.url).query).items()}

    @property
    def path(self):
        """
        Extract and return the path from the URL.

        Returns:
            str: the path of the URL.
        """
        return urlparse(self.url).path


class Response(GallonModel):
    """
    Response Object
    Interface that defines the shape of outgoing http responses
    """

    status: int = field(...)
    headers: dict = field(...)
    body: Optional[str] = field(None)

    @override
    def json(self, exclude_none: bool = True):
        """
        Convert and return the body as a JSON object.

        Returns:
            None: if body is empty.
            Json: the body as a JSON object.
        """
        if self.body is None:
            return None
        return cast(Json, loads(self.body))

    @property
    def content_type(self):
        """
        Get the value of the Content-Type header.

        Returns:
            str: the value of the Content-Type header.
        """
        return self.headers.get("Content-Type")

    @property
    def content_length(self):
        """
        Get the value of the Content-Length header.
        If Content-Length is not set, it calculates the length of the body and sets the header accordingly.

        Returns:
            str: the value of the Content-Length header.
        """
        if self.headers.get("Content-Length") is None:
            self.headers["Content-Length"] = str(len(self.body or ""))
        assert self.headers.get("Content-Length") == str(len(self.body or ""))
        return self.headers.get("Content-Length")
